import sys
import json

from apropos.goldminer.output.goldOutput import GoldOutput


class GoldJson(GoldOutput):
    "Outputs results in a list of json dictionaries."

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.as_percentage = kwargs.get("as_percentage", False)
        self.raw_values = kwargs.get("raw_values", False)
        self.out_file = kwargs.get("output_file")
        self.as_list = kwargs.get("as_list")
        self.flatten = kwargs.get("flatten", False)
        self.notfirst = False

        self.out_file_handle = sys.stdout
        if self.out_file:
            self.out_file_handle = open(self.out_file, "w")

        if self.as_list:
            self.out_file_handle.write("[")

    def maybe_comma(self):
        if self.as_list and self.notfirst:
            self.out_file_handle.write(",")
        else:
            self.notfirst = True
        self.out_file_handle.write("\n")

    def output(self, analysis_results, packet_number, tunnel_count, subscriptions):
        out_row = []
        for row in analysis_results:
            (timestamp, spi, label, value, packets) = row
            if not self.raw_values:
                value = self.calculate_confidence(label, value)

            output_struct = {
                "timestamp": float(timestamp),
                "identifiers": spi,
                "label": label,
                "value": float(value),
                "packets": int(packets),
            }

            if self.flatten:
                self.maybe_comma()
                json.dump(output_struct, self.out_file_handle)
            else:
                out_row.append(output_struct)

        if not self.flatten:
            self.maybe_comma()
            json.dump(out_row, self.out_file_handle)

    def close(self):
        if self.as_list:
            self.out_file_handle.write("\n]\n")
        self.out_file_handle.close()



"""Simplifies gold-miner output into true/false values to reduce numeric confusion"""

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
from logging import debug, info, warning, error, critical
import logging
import re
import sys
import pyfsdb
import collections


def parse_args():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description=__doc__,
        epilog="Exmaple Usage: ",
    )

    parser.add_argument(
        "-g", "--gold", default=None, type=str, help="Protocol of interest"
    )

    parser.add_argument(
        "-t", "--time-bin-size", default=20, type=float, help="Time bin size"
    )

    parser.add_argument(
        "-f",
        "--output-fsdb",
        action="store_true",
        help="Output FSDB formatted complete dataset",
    )

    parser.add_argument(
        "--log-level",
        "--ll",
        default="info",
        help="Define the logging verbosity level (debug, info, warning, error, fotal, critical).",
    )

    parser.add_argument(
        "input_file",
        type=FileType("r"),
        nargs="?",
        default=sys.stdin,
        help="Input file to read generated by gold-miner",
    )

    parser.add_argument(
        "output_file",
        type=FileType("w"),
        nargs="?",
        default=sys.stdout,
        help="Where to write results",
    )

    args = parser.parse_args()
    log_level = args.log_level.upper()
    logging.basicConfig(level=log_level, format="%(levelname)-10s:\t%(message)s")
    return args


def output_results(
    timestamp,
    current_values,
    spi_counts,
    gold_target=None,
    output_handle=sys.stdout,
    output_fsdb=False,
):
    for spi in spi_counts:
        tokens = list(current_values[spi].keys())
        if len(tokens) <= 1:
            warning(f"Insufficient gold parameters identified for {timestamp=}")
            warning(f"Only keys found: {tokens}")
            warning(f"skipping")
            return
            # exit(1)  # may be overkill for low volumes?

        best_value = -100000
        best_token = None
        output_spi = re.sub("_+", "_", re.sub("[, '()]", "_", spi))

        for rank, token in enumerate(
            sorted(tokens, key=lambda x: current_values[spi][x])
        ):
            if current_values[spi][token] > best_value:
                best_value = current_values[spi][token]
                best_token = token
            if output_fsdb and spi_counts[spi] > 0:
                output_fsdb.append(
                    [
                        timestamp,
                        output_spi,
                        token,
                        current_values[spi][token] / spi_counts[spi],
                        token == gold_target,
                        rank,
                    ]
                )

        if gold_target:
            if best_token == gold_target:
                best_token = "True"
            else:
                best_token = "False"

        if output_fsdb:
            pass  # doing this above
        else:
            output_handle.write(
                f"{timestamp},{output_spi},{best_token}, {best_value}\n"
            )


def main():
    args = parse_args()

    inh = pyfsdb.Fsdb(
        file_handle=args.input_file,
        converters={"confidence": float, "timestamp": float},
        return_type=pyfsdb.RETURN_AS_DICTIONARY,
    )

    output_fsdb = False
    if args.output_fsdb:
        output_fsdb = pyfsdb.Fsdb(out_file_handle=args.output_file)
        output_fsdb.out_column_names = [
            "timestamp",
            "spi",
            "token",
            "average_value",
            "true",
            "rank",
        ]

    timestamp_counter = 1
    next_timestamp = None
    current_values = collections.defaultdict(dict)  # indexed by spi key
    spi_counts = collections.Counter()  # indexed by spi key

    for row in inh:
        if "identifier" in row:
            spi = row["identifier"]
        else:
            spi = row["spi"]
        token = row["token"]
        confidence = row["confidence"]
        timestamp = row["timestamp"]

        if next_timestamp is None:
            next_timestamp = timestamp + args.time_bin_size
        elif timestamp >= next_timestamp:
            debug(f"values:  {current_values}")
            debug(f"counts: {spi_counts}")

            next_timestamp += args.time_bin_size

            # report results
            output_results(
                timestamp_counter * args.time_bin_size,
                current_values,
                spi_counts,
                gold_target=args.gold,
                output_handle=args.output_file,
                output_fsdb=output_fsdb,
            )

            # reset these
            current_values = collections.defaultdict(dict)
            spi_counts = collections.Counter()
            timestamp_counter += 1

        if token not in current_values[spi]:
            current_values[spi][token] = 0.0
        current_values[spi][token] += confidence
        spi_counts[spi] += 1


if __name__ == "__main__":
    main()



from logging import debug, warning
import collections
import pyfsdb

from apropos.goldminer.output.goldOutput import GoldOutput


class GoldFsdb(GoldOutput):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.as_percentage = kwargs.get("as_percentage", False)
        self.raw_values = kwargs.get("raw_values", False)
        self._save_values = kwargs.get("save_values", None)

        # delete keys we can't pass onward to FSDB
        for key in ["as_percentage", "save_values", "raw_values", "parameter_file"]:
            if key in kwargs:
                del kwargs[key]

        self.outh = pyfsdb.Fsdb(**kwargs)
        self.outh.out_column_names = [
            "timestamp",
            "identifier",
            "token",
            "confidence",
            "total",
        ]

        if self._save_values:
            self._save_values_storage = collections.defaultdict(dict)

    def output(self, analysis_results, packet_number, tunnel_count, subscriptions):
        for row in analysis_results:
            (timestamp, identifier, label, value, packets) = row
            if not self.raw_values:
                value = self.calculate_confidence(label, value)
            self.outh.append([timestamp, identifier, label, value, packets])

            # save the last label seen
            # TODO: save min/max/avg/binned/etc?
            if self._save_values:
                self._save_values_storage[identifier][label] = value

    @property
    def save_values(self):
        return self._save_values_storage

    def close(self):
        self.outh.close()



from logging import warning, debug, info

from apropos.goldminer.output.goldOutput import GoldOutput


class GoldHighLowWatermark(GoldOutput):
    """Truncates an output generator to just when values pass a high/low
    watermark.  When an incoming value goes beyond the *high* mark, a
    (single) report is made.  When the value drops below the *low*
    mark afterward, another report is made indicating it dropped below
    again.  All other reports are suppressed, indicating the last
    reported state is still current.

    The *output* option is required and must point to a GoldOutput instance
    of some kind.

    Note that outputs with more than one result will be tracked one at
    a time, with each identifier indicating a separate reporting
    system.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.high = kwargs.get("high", 0.0)
        self.low = kwargs.get("low", 1.0)
        self._output = kwargs.get("output")
        self.reported = {}

        if not self._output:
            raise ValueError("the output option is required for GoldHighLowWatermark")

    def output(self, analysis_results, packet_number, tunnel_count, subscriptions):
        for row in analysis_results:
            identifier = row[1]
            value = row[3]
            if identifier not in self.reported:
                self.reported[identifier] = False
            if not self.reported[identifier] and value >= self.high:
                self._output.output([row], packet_number, tunnel_count, subscriptions)
                self.reported[identifier] = True
            elif self.reported[identifier] and value < self.low:
                self._output.output([row], packet_number, tunnel_count, subscriptions)
                self.reported[identifier] = False

    def close(self):
        self._output.close()

    def end_section(self):
        self._output.end_section()

    def append(self, data):
        self._output.append(data)

    def calculate_confidence(self, label, value):
        return self._output.calculate_confidence(label, value)



import logging


class GoldOutput:
    def __init__(self, **kwargs):
        self.trained_parameters = {}

        if "parameter_file" in kwargs and kwargs["parameter_file"]:
            import pyfsdb

            f = pyfsdb.Fsdb(
                kwargs["parameter_file"],
                return_type=pyfsdb.RETURN_AS_DICTIONARY,
                converters={"low_val": float, "high_val": float},
            )
            for row in f:
                self.trained_parameters[row["key"]] = [row["low_val"], row["high_val"]]
                logging.debug(self.trained_parameters)
        else:
            # logging.warning(f"No parameter file passed to {self}")
            pass

    def calculate_confidence(self, label, value):
        try:
            (specific_low, specific_high) = self.trained_parameters.get(
                label, [0.0, 1.0]
            )
            fraction = max(
                min((value - specific_low) / (specific_high - specific_low), 1.0), 0.0
            )
        except Exception as e:
            logging.error(f"failed confidence gathering: {e}")
            return -1

        return fraction

    def close(self):
        pass

    def end_section(self):
        pass

    def append(self, data):
        pass



from logging import debug, warning
import collections
import pyfsdb

from apropos.goldminer.output.goldOutput import GoldOutput


class GoldNoOp(GoldOutput):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def output(self, analysis_results, packet_number, tunnel_count, subscriptions):
        pass



import unittest


class TestPickAxeMath(unittest.TestCase):
    def test_difference_function(self):
        from apropos.goldminer.pickaxe.pickaxe import PickAxe

        a = PickAxe()

        # measured, dirty_min, gold_value, dirty_max, expected
        test_cases = [
            [0.2, 0.4, 0.3, 0.8, 0.0],  # measured < gold < min < max
            [0.8, 0.1, 0.7, 0.6, 0.0],  # min < max < gold < measured
            # gold(.4) < measured(.5) < min(.8) < max(.9) = (.8-.5)/(.8-.4) = .75
            [0.5, 0.8, 0.4, 0.9, 0.25],
            # gold(.6) > measured(.5) > max(.2) > min(.1) = (.8-.5)/(.8-.4) = .75
            [0.5, 0.1, 0.6, 0.2, 0.25],
            #
            # measured is above max and gold is below
            # min(.1) < gold(.4) < max(.8) < measured(.9)
            [0.9, 0.1, 0.4, 0.8, 1.0],
            # gold(.1) < min(.2) <  < max(.8) < measured(.9)
            [0.9, 0.2, 0.1, 0.8, 1.0],
            #
            # measured is below min and gold above
            # measured(.1) < min(.2) < gold(.4) < max(.8)
            [0.1, 0.2, 0.4, 0.8, 1.0],
            # measured(.1) < min(.2) < max(.8) < gold(.9)
            [0.1, 0.2, 0.9, 0.8, 1.0],
            #
            # measured between min and gold, with different values:
            # min(.2) < measured(.3) < gold(.6) < max(.8)
            [0.3, 0.2, 0.6, 0.8, 0.75],
            # min(.2) < measured(.5) < gold(.6) < max(.8)
            [0.5, 0.2, 0.6, 0.8, 0.25],
            #
            # measured between max and gold, with different values:
            # min(.1) < gold(.2) < measured(.5) < max(.6)
            [0.5, 0.1, 0.2, 0.6, 0.75],
            # min(.1) < gold(.2) < measured(.3) < max(.6)
            [0.3, 0.1, 0.2, 0.6, 0.25],
            #
            # min(.3) < gold(.4) == measured(.4) == max(.4)
            # gold == dirty_max (divide by zero error - return )
            [0.4, 0.3, 0.4, 0.4, 0.0],
        ]

        for n, test in enumerate(test_cases):
            result = round(a.distance_by_dirty_comparison(*test[0:4], 0, 0), 3)
            self.assertEqual(result, test[4], f"testing case {n}")



#!/usr/bin/python3

import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import logging
from logging import debug, info, warning, error, critical
from apropos.goldminer.trainer import GoldMineTrainer
from apropos.goldminer.trainer.widthtrainer import GoldMineWidthTrainer


def parse_args():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description=__doc__,
        epilog="Exmaple Usage: " + "pcap-to-fsdb.py -o output.fsdb input.pcap",
    )

    parser.add_argument(
        "-o",
        "--output",
        default=sys.stdout,
        type=FileType("w"),
        help="Where to write the output file (else stdout)",
    )

    parser.add_argument(
        "-a",
        "--addresses",
        default=None,
        type=str,
        nargs="*",
        help="Only look at packets from these addresses",
    )

    parser.add_argument(
        "-f",
        "--filter",
        default=None,
        help="Use a pcap filter to filter packets",
    )

    parser.add_argument(
        "-T",
        "--no-timing",
        action="store_true",
        help="Don't do timing analysis (expensive in memory)",
    )

    parser.add_argument(
        "--log-level",
        "--ll",
        default="info",
        help="Define the logging verbosity level (debug, info, warning, error, fotal, critical).",
    )

    parser.add_argument(
        "-W",
        "--width-trainer",
        action="store_true",
        help="Use the [beta] width trainer",
    )

    parser.add_argument(
        "pcap_files", nargs="+", help="Where to read the pcap file(s) from"
    )

    args = parser.parse_args()

    log_level = args.log_level.upper()
    logging.basicConfig(level=log_level, format="%(levelname)-10s:\t%(message)s")

    return args


def main():
    """main routine for arg wrapping and calling analysis"""
    args = parse_args()

    if args.width_trainer:
        trainer = GoldMineWidthTrainer(addresses=args.addresses)
    else:
        trainer = GoldMineTrainer(
            do_timing_analysis=(not args.no_timing), addresses=args.addresses
        )

    trainer.analyze_files(args.pcap_files, pkt_filter=args.filter)
    trainer.generate_results()
    trainer.output_to_fsdb(out_file_handle=args.output)


if __name__ == "__main__":
    main()



from apropos.goldminer.output.goldOutput import GoldOutput


class GoldTextDump(GoldOutput):
    def __init__(self, seperator="\n"):
        self.seperator = seperator

    def output(self, analysis_results, packet_number, tunnel_count, subscriptions):
        outputs = []
        strformat = "n={}\tt={}\tspi={}\ttoken={}:\t{}"
        outputs = [
            strformat.format(z, v, w, x, y) for (v, w, x, y, z) in analysis_results
        ]
        output = self.seperator.join(outputs)
        print(output, end=self.seperator)



import sys
import collections
import numpy as np
from logging import debug, info, warning, error, critical
from apropos.goldminer.trainer import GoldMineBase


class GoldMineWidthTrainer(GoldMineBase):
    def __init__(self, addresses=None, max_width=10000, max_pkt_size=1500):
        super().__init__(addresses)
        self.packet_data = collections.defaultdict(collections.Counter)

        # create the data keys to use
        self.packet_keys = ["size_count"]
        self.percent_keys = ["percent_size_count"]

        self.all_keys = self.packet_keys + self.percent_keys

        self.storage = np.zeros((max_pkt_size, max_width), dtype=bool)

    def analyze_packet(self, pkt, count):
        """Analyzes a packet to pull out features."""
        # count the packet sizes
        self.storage[pkt.len][count] = True

    def generate_results(self):
        """Normalize the counts to fractions"""
        import pdb

        pdb.set_trace()
        self.total_counts = collections.Counter()

        # calculate totals
        for data_key in self.packet_data:
            for key in self.packet_keys:
                self.total_counts["percent_" + key] += int(
                    self.packet_data[data_key][key]
                )

        # now divide:
        for data_key in self.packet_data:
            for key in self.packet_keys:
                self.packet_data[data_key]["percent_" + key] = float(
                    (self.packet_data[data_key][key])
                    / float(self.total_counts["percent_" + key])
                )

        if self.do_timing_analysis:
            self.analyze_packet_size_delays()

    def analyze_packet_size_delays(self):
        import numpy  # only required if doing time delays

        for size in self.packet_size_delays:
            arr = numpy.array(self.packet_size_delays[size])
            self.packet_data[size]["timing_mean"] = arr.mean()
            self.packet_data[size]["timing_median"] = numpy.median(arr)
            self.packet_data[size]["timing_std"] = numpy.std(arr)
            self.packet_data[size]["timing_max"] = arr.max()

            if len(self.packet_size_timestamps[size]) > 1:
                deltas = []
                for n in range(1, len(self.packet_size_timestamps[size])):
                    deltas.append(
                        self.packet_size_timestamps[size][n]
                        - self.packet_size_timestamps[size][n - 1]
                    )
                arr = numpy.array(deltas).astype(float)
                self.packet_data[size]["ipa_mean"] = arr.mean()
                self.packet_data[size]["ipa_min"] = arr.min()
                self.packet_data[size]["ipa_max"] = arr.max()
                self.packet_data[size]["ipa_std"] = numpy.std(arr)

                quantiles = numpy.quantile(arr, [0.25, 0.5, 0.75])
                self.packet_data[size]["ipa_q1"] = quantiles[0]
                self.packet_data[size]["ipa_median"] = quantiles[1]
                self.packet_data[size]["ipa_q3"] = quantiles[2]

            else:
                self.packet_data[size]["ipa_mean"] = 0
                self.packet_data[size]["ipa_min"] = 0
                self.packet_data[size]["ipa_max"] = 0
                self.packet_data[size]["ipa_std"] = 0
                self.packet_data[size]["ipa_q1"] = 0
                self.packet_data[size]["ipa_median"] = 0
                self.packet_data[size]["ipa_q3"] = 0

    def output_to_fsdb(self, out_file_handle):
        import pyfsdb

        outh = pyfsdb.Fsdb(out_file_handle=out_file_handle)
        outh.out_column_names = ["key"] + self.all_keys

        for data_key in self.packet_data:
            outh.append(
                [data_key] + [self.packet_data[data_key][x] for x in self.all_keys]
            )
        outh.close()



import sys
import collections
from logging import debug, info, warning, error, critical
import dpkt


class GoldMineBase:
    def __init__(self, do_timing_analysis=True, addresses=None):
        self.address_filters = addresses
        self.reset()

    def reset(self):
        self.n = -1  # since we increment immediately, start at -1 to get to zero
        self.last_size = 0

    def analyze(self, timestamp, packet):
        # discard non-ips
        try:
            self.n += 1

            eth = dpkt.ethernet.Ethernet(packet)
            ip = eth.data
            if not isinstance(ip, dpkt.ip.IP):
                warning("A non-IP packet was received")

            # ignore things out of the address block we want to inspect
            if self.address_filters:
                for address in self.address_filters:
                    if "{}.{}.{}.{}".format(*ip.src).startswith(address):
                        break
                else:
                    return  # skip this packet

            self.analyze_packet(timestamp, ip, self.last_size + self.n)
            self.last_size += self.n + 1

        except Exception:
            warning(f"packet #{self.n} could not be parsed")

    def analyze_files(self, pcap_files: list, count: int = 0, pkt_filter: str = None):
        self.reset()
        for n, pcap_file in enumerate(pcap_files):
            info(f"ANALYZING: {pcap_file} with {pkt_filter=} {count=}")

            # open the pcap file
            pcap = dpkt.pcap.Reader(open(pcap_file, "rb"))

            # optionally set the filter
            if pkt_filter:
                pcap.setfilter(pkt_filter)

            pcap.dispatch(0, self.analyze)


class GoldMineTrainer(GoldMineBase):
    def __init__(self, do_timing_analysis=True, addresses=None):
        super().__init__(addresses)
        self.packet_data = collections.defaultdict(collections.Counter)
        self.timing_keys = ["timing_mean", "timing_max", "timing_median", "timing_std"]
        self.ipa_keys = [
            "ipa_mean",
            "ipa_median",
            "ipa_q1",
            "ipa_q3",
            "ipa_min",
            "ipa_max",
            "ipa_std",
        ]
        self.first_packet_time = None
        self.last_packet_time = None
        self.packet_size_delays = {}
        self.packet_size_timestamps = {}
        self.do_timing_analysis = do_timing_analysis

        # create the data keys to use
        if do_timing_analysis:
            self.packet_keys = ["ipa_time", "size_count"]
        else:
            self.packet_keys = ["size_count"]
        self.percent_keys = ["percent_" + x for x in self.packet_keys]  # transform

        self.all_keys = self.packet_keys + self.percent_keys
        if do_timing_analysis:
            self.all_keys += self.timing_keys + self.ipa_keys

    def analyze_packet(self, timestamp, packet, count):
        """Analyzes a packet to pull out features."""
        global last_packet_time
        global first_packet_time

        # get the length
        packet_len = len(packet)

        # count the packet sizes
        self.packet_data[packet_len]["size_count"] += 1

        if self.do_timing_analysis:
            # calculate the time since the last packet
            # (this turns out to be pretty much always 0)
            this_time = timestamp
            if self.last_packet_time:
                self.packet_data[int(this_time - self.last_packet_time)][
                    "ipa_time"
                ] += 1

            # remember this timestamp
            self.last_packet_time = this_time

            # create a list of real times since this last *size* was seen
            if not self.first_packet_time:
                self.first_packet_time = this_time

            if len(packet) not in self.packet_size_delays:
                self.packet_size_delays[packet_len] = [0.0]
                self.packet_size_timestamps[packet_len] = [this_time]
            else:
                self.packet_size_delays[packet_len].append(
                    float(this_time - self.first_packet_time)
                )
                self.packet_size_timestamps[packet_len].append(this_time)

            self.first_packet_time = this_time

    def generate_results(self):
        """Normalize the counts to fractions"""
        self.total_counts = collections.Counter()

        # calculate totals
        for data_key in self.packet_data:
            for key in self.packet_keys:
                self.total_counts["percent_" + key] += int(
                    self.packet_data[data_key][key]
                )

        # now divide:
        for data_key in self.packet_data:
            for key in self.packet_keys:
                self.packet_data[data_key]["percent_" + key] = float(
                    (self.packet_data[data_key][key])
                    / float(self.total_counts["percent_" + key])
                )

        if self.do_timing_analysis:
            self.analyze_packet_size_delays()

    def analyze_packet_size_delays(self):
        import numpy  # only required if doing time delays

        for size in self.packet_size_delays:
            arr = numpy.array(self.packet_size_delays[size])
            self.packet_data[size]["timing_mean"] = arr.mean()
            self.packet_data[size]["timing_median"] = numpy.median(arr)
            self.packet_data[size]["timing_std"] = numpy.std(arr)
            self.packet_data[size]["timing_max"] = arr.max()

            if len(self.packet_size_timestamps[size]) > 1:
                deltas = []
                for n in range(1, len(self.packet_size_timestamps[size])):
                    deltas.append(
                        self.packet_size_timestamps[size][n]
                        - self.packet_size_timestamps[size][n - 1]
                    )
                arr = numpy.array(deltas).astype(float)
                self.packet_data[size]["ipa_mean"] = arr.mean()
                self.packet_data[size]["ipa_min"] = arr.min()
                self.packet_data[size]["ipa_max"] = arr.max()
                self.packet_data[size]["ipa_std"] = numpy.std(arr)

                quantiles = numpy.quantile(arr, [0.25, 0.5, 0.75])
                self.packet_data[size]["ipa_q1"] = quantiles[0]
                self.packet_data[size]["ipa_median"] = quantiles[1]
                self.packet_data[size]["ipa_q3"] = quantiles[2]

            else:
                self.packet_data[size]["ipa_mean"] = 0
                self.packet_data[size]["ipa_min"] = 0
                self.packet_data[size]["ipa_max"] = 0
                self.packet_data[size]["ipa_std"] = 0
                self.packet_data[size]["ipa_q1"] = 0
                self.packet_data[size]["ipa_median"] = 0
                self.packet_data[size]["ipa_q3"] = 0

    def output_to_fsdb(self, out_file_handle):
        import pyfsdb

        outh = pyfsdb.Fsdb(out_file_handle=out_file_handle)
        outh.out_column_names = ["key"] + self.all_keys

        for data_key in self.packet_data:
            outh.append(
                [data_key] + [self.packet_data[data_key][x] for x in self.all_keys]
            )
        outh.close()

    @property
    def results(self):
        data = []
        for data_key in self.packet_data:
            data.append(
                [data_key] + [self.packet_data[data_key][x] for x in self.all_keys]
            )
        return data



import unittest


class TestLoads(unittest.TestCase):
    def test_pickaxe(self):
        from apropos.goldminer.pickaxe.pickaxe import PickAxe

        a = PickAxe()

    def test_trainer(self):
        from apropos.goldminer.trainer import GoldMineTrainer

        t = GoldMineTrainer()

        from apropos.goldminer.trainer.widthtrainer import GoldMineWidthTrainer

        t = GoldMineWidthTrainer()



import curses

from apropos.goldminer.output.goldOutput import GoldOutput


class GoldCurses(GoldOutput):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.win = curses.initscr()
        curses.start_color()
        curses.init_pair(1, curses.COLOR_RED, curses.COLOR_WHITE)
        curses.init_pair(2, curses.COLOR_YELLOW, curses.COLOR_WHITE)
        curses.init_pair(3, curses.COLOR_GREEN, curses.COLOR_WHITE)

        self.win.move(0, 1)
        self.win.addstr("spi")

        self.win.move(0, 12)
        self.win.addstr("pkts")

        self.win.move(0, 20)
        self.win.addstr("label")

        self.win.move(0, 30)
        self.win.addstr("confidence")

        # XXX hard coded
        self.bar_length = 20
        self.as_percentage = kwargs.get("as_percentage")

    def close(self):
        import time

        time.sleep(1)
        curses.endwin()

    def output(self, analysis_results, packet_number, tunnel_count, subscriptions):
        lastspi = ""
        for n0, output in enumerate(analysis_results):
            (timestamp, spi, label, value, packets) = output

            n = n0 + 1
            g = n0 + 23

            # plot the label
            # XXX: only need to do this once really
            if spi != lastspi:
                lastspi = spi
                try:
                    spi = hex(spi)
                except Exception:
                    spi = "None"

                self.win.move(n, 1)
                self.win.addstr(spi)

                self.win.move(n, 12)
                self.win.addstr(str(packets))

            self.win.move(n, 20)
            self.win.addstr(label[-9:])

            # calculate the fraction of traffic
            # import sys
            # sys.stderr.write(label + "\n")
            if label not in self.trained_parameters:
                self.win.move(g, 1)
                self.win.addstr("Failed to find parameters for {}".format(label))
            else:
                self.win.move(g, 1)
                self.win.addstr(
                    "Gold: {} : {} - {}".format(label, *self.trained_parameters[label])
                )

            # turn the result value into a confidence score
            fraction = self.calculate_confidence(label, value)

            # plot the raw value
            self.win.move(n, 30)
            if self.as_percentage:
                self.win.addstr(str(int(100.0 * fraction)) + "   ")
            else:
                self.win.addstr(str(value)[0:23])

            # plot the bar
            self.win.move(n, 55)

            color = 2
            if fraction < 0.25:
                color = 1
            elif fraction > 0.75:
                color = 3

            left_count = int(self.bar_length * fraction)
            right_count = self.bar_length - left_count

            self.win.addstr(
                ("=" * left_count) + ("_" * right_count), curses.color_pair(color)
            )

            # self.win.addstr(str(left_count))

        self.win.move(20, 1)
        self.win.addstr(f"Packets seen: {packet_number}")
        self.win.move(21, 1)
        self.win.addstr(f"Tunnels seen: {tunnel_count}")
        self.win.move(22, 1)
        self.win.addstr("Outputs seen: {}".format(len(analysis_results)))

        self.win.refresh()



from collections import Counter, defaultdict
from logging import debug, info, warning, error, critical
import dpkt
import numpy as np
import time
import pyfsdb
import pprint
from apropos.goldminer.output.noop import GoldNoOp

#
# Analyzer
#
last_packet_time = {}


class PickAxe:
    def __init__(
        self,
        gold_profiles: list = [],
        training_data_file: str = None,
        training_key: str = "e_pkt_len",
        support_windowing: bool = False,
        threshold_file=None,
        algorithm="comparison",
        label_map=None,
        report_every_n=None,
        three_tuple_only=False,
    ):

        # dictionary of combinations to search for
        #   each entry is a sub-dict with gold/dirty names with accumulated protocol information
        self.gold_mines = {}
        self._gold_profiles = gold_profiles
        self._training_data_file = training_data_file
        self.training_key = training_key
        self.threshold_file = threshold_file
        self._label_map = label_map
        self.skip_packets = 0
        self.missing_in_profile_penalty = 1.0  # TODO: configurable
        self.missing_in_tunnel_penalty = 1.0  # TODO: configurable
        self._output = GoldNoOp()
        self.report_every_n = report_every_n
        self.three_tuple_only = three_tuple_only

        # A queue and server for us when connected to a GRPC server
        self.server = None
        self.subscriptions = {}
        self.packet_count = 0
        self.tunnels = defaultdict(dict)

        self._outputs = None
        self.all_keys = []

        # starting keep length
        self.starting_length = 30000

        if self.gold_profiles:
            self.read_percents()

        # create a (sparse) counting map
        self.maximum_identifiers = 200
        self.identifier_indexes = {}
        # we use 1550 just to account for some labeling that gets bigger than 1500
        self.support_windowing = support_windowing
        self.size_array = None
        if self.support_windowing:
            # TODO need to deal with reallocating this
            self.size_array = np.zeros(
                (self.maximum_identifiers, 1550, self.starting_length), dtype=bool
            )
        self.tunnel_packets = Counter()

        # remember the algorithm functions
        self._algorithms = {
            "lms": self.distance_by_lms,
            "linear": self.distance_by_linear,
            "comparison-wide": self.distance_by_dirty_comparison,
            "comparison": self.distance_by_dirty_comparison2,
        }
        self.calculate_distance = self.distance_by_dirty_comparison
        self.algorithm = algorithm

    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, value):
        self._output = value

    @property
    def gold_profiles(self):
        return self._gold_profiles

    @gold_profiles.setter
    def gold_profiles(self, value):
        self._gold_profiles = value
        self.read_percents()

    @property
    def algorithm(self):
        "The current algorithm to be used when processing data"
        for key in self._algorithms:
            if self._algorithms[key] == self.calculate_distance:
                return key
        error("no algorithm set")
        return "unknown"

    @algorithm.setter
    def algorithm(self, value):
        "Set the current algorithm to be used when processing data"
        if value not in self._algorithms:
            error(f"unknown algorithm passed to {self}: {value}")
            raise ValueError(f"unknown algorithm: {value}")
        self.calculate_distance = self._algorithms[value]

    @property
    def algorithms(self):
        "The algorithms available to be used"
        return list(self._algorithms.keys())

    @property
    def label_map(self):
        return self._label_map

    @label_map.setter
    def label_map(self, newval):
        self._label_map = newval

    @property
    def training_data_file(self):
        "The training data file being used"
        return self._training_data_file

    @training_data_file.setter
    def training_data_file(self, value):
        "The training data file to use"
        self._training_data_file = value
        self.read_percents()

    def process_pcap(
        self,
        pcap_file: str,
        pkt_filter: str = None,
        max_packets: int = 0,
        search_window_length: int = 0,
        dig_for_gold: bool = True,
        skip_packets: int = 0,
    ) -> None:

        self.skip_packets = skip_packets
        pkt_count = max_packets + self.skip_packets
        info(
            f"ANALYZING: {pcap_file} with {pkt_filter=} {max_packets=} {skip_packets=}"
        )

        # open the pcap file
        pcap = dpkt.pcap.Reader(open(pcap_file, "rb"))

        # optionally set the filter
        if pkt_filter:
            pcap.setfilter(pkt_filter)

        pcap.dispatch(pkt_count, self.analyze_packet)

        return (self.identifier_indexes, self.tunnel_packets, self.size_array)

    def process_interface(self, interface: str, packet_filter: str = None):

        debug(f"  sniffing({interface})")
        sniff(prn=self.analyze_packet, iface=interface, filter=packet_filter, store=0)

    def find_count_for_time(
        self, array: list, since_timestamp: int, max_len: int, pkt_size: int
    ) -> int:
        """Find the count number of entries in an array of recorded timestamps
        for each packet size since a *since_timestamp* timestamp.

        TODO needs rethinking:
        If since_timestamp is a dictionary, we use the size as an index to
        search instead for different time lengths per size.

        TODO: this could be sped up (if nothing else, remember the point of the last time found)
        """
        if isinstance(since_timestamp, dict):
            timestamp = since_timestamp.get(pkt_size, 2147483647)
            count = max_len - array[:max_len].searchsorted(timestamp)
        else:
            count = max_len - array[:max_len].searchsorted(since_timestamp)
        return count

    def read_percents(self):
        filename = self.training_data_file
        all_gold_keys = self.gold_profiles

        fh = pyfsdb.Fsdb(filename, return_type=pyfsdb.RETURN_AS_DICTIONARY)

        # all data from the percentage file
        all_data = []

        column_names = fh.column_names

        if all_gold_keys == []:
            # assume they want all gold
            all_gold_keys = [x for x in column_names if x != self.training_key]
            self.gold_profiles = all_gold_keys
            debug(f"assuming all keys: {self.gold_profiles}")

        # reset
        self.gold_mines = {}

        # if we weren't passed dirty keys, create them
        have_dirty = True
        if len(self.all_keys) == 0:
            have_dirty = False

        # load all the TRAINING data into an array
        all_data = fh.get_all()

        # for each gold profile, build a list
        for gold_profile in all_gold_keys:
            # load all the good data into self.gold_percents per key
            self.gold_percents = {}

            # the list of keys we want to find
            gold_keys = gold_profile.split(",")

            # the list of OTHER columns in the data
            dirty_keys = []

            # for every column name, place it in a gold list or dirty list of keys
            for column in column_names:
                if column in gold_keys:
                    self.gold_percents[column] = {}
                elif not have_dirty and column != self.training_key:
                    dirty_keys.append(column)
                elif have_dirty and column in self.all_keys:
                    dirty_keys.append(column)

            # load all the dirty data as an average into percents
            percents = {}

            # loop through all data and place it in
            for row in all_data:
                try:
                    size = int(row[self.training_key])
                except Exception:
                    error(
                        f"failed to parse this training key as an integer: '{row[self.training_key]}'"
                    )
                    continue  # failed to parse to an int -- ignore this

                # collect the gold key values first
                gold_max = -1
                gold_min = 1000
                for key in gold_keys:
                    try:
                        value = float(row[key])
                    except Exception:
                        value = 0.0

                    if key not in self.gold_percents:
                        error(f"gold profile '{key}' not found in data")
                        continue

                    self.gold_percents[key][size] = value

                    gold_max = max(self.gold_percents[key][size], gold_max)
                    gold_min = min(self.gold_percents[key][size], gold_min)

                if gold_max == 0.0 and gold_min == 0.0:
                    continue

                # TODO: deal with len(self.gold_percents[key].values())  == 0
                # how do we decrease the probability of gold given other things?  or do we?
                # probably don't need to, as we assume other things in the stream is ok

                # TODO: we're not handling multiple AND/OR golds right now - just averaging
                gold_value = (gold_max - gold_min) / 2 + gold_min

                # now collect the min/max above the measured gold_min/max
                dirty_max = -1
                dirty_min = 10000
                dirty_range_max = 10000
                dirty_range_min = -1

                for key in dirty_keys:
                    try:
                        value = float(row[key])

                        # XXX TODO FIX: should be min/max nearest gold value

                        if value >= gold_value:
                            # take the lowest value above the gold value
                            # thus dirty_max is the min [sic] above gold
                            dirty_range_max = min(value, dirty_range_max)

                        if value <= gold_value:
                            # take the highest value above the gold value
                            # thus dirty_max is the max [sic] above gold
                            dirty_range_min = max(value, dirty_range_min)

                        dirty_max = max(value, dirty_max)
                        dirty_min = min(value, dirty_min)
                    except Exception:
                        pass  # value was not a float (likely '' or '-')

                # remember our dirty high/low values and the gold_value per size
                percents[size] = (
                    dirty_min,
                    gold_value,
                    dirty_max,
                    dirty_range_max,
                    dirty_range_min,
                )

                debug(f"{size}: {percents[size]}  ({gold_max}, {gold_min})")

            # we're done build the trained results for this profile
            self.gold_mines[gold_profile] = percents

            debug("gold mines:")
            debug(pprint.pformat(self.gold_mines))

    def calculate_distance(
        self,
        measured,
        dirty_min,
        gold_value,
        dirty_max,
        dirty_range_max,
        dirty_range_min,
    ):
        raise ValueError("algorithm unset")

    def distance_by_lms(
        self,
        measured,
        dirty_min,
        gold_value,
        dirty_max,
        dirty_range_max,
        dirty_range_min,
    ):
        diff = measured - gold_value
        return diff * diff

    def distance_by_linear(
        self,
        measured,
        dirty_min,
        gold_value,
        dirty_max,
        dirty_range_max,
        dirty_range_min,
    ):
        return abs(measured - gold_value)

    def distance_by_dirty_comparison(
        self,
        measured,
        dirty_min,
        gold_value,
        dirty_max,
        dirty_range_max,
        dirty_range_min,
    ):
        """Determine how far away a measured value is between the target gold
        and the nearest dirty marker(s).  Positive values are
        considered better.
        """
        difference = None
        if dirty_min == 10000 or dirty_max == -1:
            # TODO: return N/N+1 instead
            # (to accomplish this we need to remember N, which we don't yet)
            difference = 1  # no dirty records for this point .
            if not (dirty_min == 10000 and dirty_max == -1):
                error("bad math -- min/max are not both extreme indicating missing")
        elif measured < dirty_min:
            if gold_value < dirty_min:
                if measured < gold_value:
                    # measured < gold < dirty_min
                    difference = 1
                else:
                    # gold < measured < dirty_min
                    # return fraction of distance from measured to dirty
                    # (w / 1 being better)
                    difference = (dirty_min - measured) / (dirty_min - gold_value)
            else:
                # measured < dirty_min and gold > dirty_min
                difference = 0
        elif measured > dirty_max:
            if gold_value > dirty_max:
                if measured > gold_value:
                    # measured > gold > dirty_max
                    difference = 1
                else:
                    # gold > measured > dirty_max
                    # result = fraction of distance from dirty to measured
                    difference = (measured - dirty_max) / (gold_value - dirty_max)
            else:
                if measured > gold_value:  ## TODO: wrong -- gold < max here
                    # measured > gold > max
                    difference = 0
                else:
                    # gold > measured > max
                    difference = (dirty_max - measured) / (dirty_max - gold_value)
                    # measured > max
        else:
            # dirty_min < measured < dirty_max
            if gold_value < dirty_min or gold_value > dirty_max:
                difference = 0
            elif gold_value > measured:
                # min < measured < gold < max
                # return fraction of distance close to min from measured
                difference = (measured - dirty_min) / (gold_value - dirty_min)
            elif dirty_max == gold_value:
                # div/0 error if we did the below, so return 0 (not good)
                return 0
            else:
                # min < gold < measured < max
                # return fraction of distance close to max from measured
                difference = (dirty_max - measured) / (dirty_max - gold_value)

        if difference is None:
            error("OH NOOOOOOOOOOOOOOOOOOOO - empty result means we missed a case")
            error(
                f"r={difference}, gv={gold_value}, dmin={dirty_min}, dmax={dirty_max}, m={measured}"
            )
            difference = 0.5

        return 1 - difference  # we want 0 is better, so invert this

    def distance_by_dirty_comparison2(
        self,
        measured,
        dirty_min,
        gold_value,
        dirty_max,
        dirty_range_max,
        dirty_range_min,
    ):
        """Determine how far away a measured value is between the target gold
        and the nearest dirty marker(s).  Positive values are
        considered better.
        """
        dirty_min = dirty_range_min
        dirty_max = dirty_range_max

        difference = None
        if dirty_min == 10000 or dirty_max == -1:
            # TODO: return N/N+1 instead
            # (to accomplish this we need to remember N, which we don't yet)
            difference = 1  # no dirty records for this point .
            if not (dirty_min == 10000 and dirty_max == -1):
                error("bad math -- min/max are not both extreme indicating missing")
        elif measured < dirty_min:
            if gold_value < dirty_min:
                if measured < gold_value:
                    # measured < gold < dirty_min
                    difference = 1
                else:
                    # gold < measured < dirty_min
                    # return fraction of distance from measured to dirty
                    # (w / 1 being better)
                    difference = (dirty_min - measured) / (dirty_min - gold_value)
            else:
                # measured < dirty_min and gold > dirty_min
                difference = 0
        elif measured > dirty_max:
            if gold_value > dirty_max:
                if measured > gold_value:
                    # measured > gold > dirty_max
                    difference = 1
                else:
                    # gold > measured > dirty_max
                    # result = fraction of distance from dirty to measured
                    difference = (measured - dirty_max) / (gold_value - dirty_max)
            else:
                if measured > gold_value:  ## TODO: wrong -- gold < max here
                    # measured > gold > max
                    difference = 0
                else:
                    # gold > measured > max
                    difference = (dirty_max - measured) / (dirty_max - gold_value)
                    # measured > max
        else:
            # dirty_min < measured < dirty_max
            if gold_value < dirty_min or gold_value > dirty_max:
                difference = 0
            elif gold_value > measured:
                # min < measured < gold < max
                # return fraction of distance close to min from measured
                difference = (measured - dirty_min) / (gold_value - dirty_min)
            elif dirty_max == gold_value:
                # div/0 error if we did the below, so return 0 (not good)
                return 0
            else:
                # min < gold < measured < max
                # return fraction of distance close to max from measured
                difference = (dirty_max - measured) / (dirty_max - gold_value)

        if difference is None:
            error("OH NOOOOOOOOOOOOOOOOOOOO - empty result means we missed a case")
            error(
                f"r={difference}, gv={gold_value}, dmin={dirty_min}, dmax={dirty_max}, m={measured}"
            )
            difference = 0.5

        return 1 - difference  # we want 0 is better, so invert this

    def dig_4it(self, timestamp: int, window_length: int = 0, identifier=None) -> None:
        """Given a sequence of packets in various sizes, determine if anything
        is one of the hopeful packet types given their feature counts."""
        output_results = {}
        debug("----")

        beginning_timestamp = 0
        if window_length:
            if isinstance(window_length, dict):
                beginning_timestamp = window_length
            else:
                beginning_timestamp = float(timestamp) - window_length

        tunnel_list = self.tunnels

        # ignore all other tunnels but the ones marked by a passed identifier
        if identifier:
            tunnel_list = [identifier]

        # examine each tunnel we've seen or was requested to study
        for identifier in tunnel_list:

            # calculate the total number of packets seen from all tunnels
            tunnel = self.tunnels[identifier]
            tunnel_packets = self.tunnel_packets[identifier]

            # loops over the entire set of seen packets to date by pkt_size
            # mine == named search key
            for protocol_label in self.gold_mines:
                training_percents = self.gold_mines[protocol_label]
                difference_total = 0.0
                number_of_sizes = 0

                # for every protocol we are searching for,
                # evaluate it for this tunnel's packet sizes
                for pkt_size in tunnel:
                    number_of_sizes += 1

                    if pkt_size not in training_percents:
                        # since this packet size has never been seen in
                        # our target "gold" protocol, it's a sign it may
                        # be something different.  Punish it accordingly.
                        difference_total += self.missing_in_profile_penalty
                        # debug(f"unknown packet size seen: {pkt_size}")
                        continue  # skip the rest of the calculations

                    # the current percentage seen for this size
                    # measured = tunnel[pkt_size]['length'] / identifier_packets
                    measured = (
                        # TODO: this complexity isn't needed unless
                        # windowing is used should be able to just
                        # take the total length all the time for
                        # never-ending
                        self.find_count_for_time(
                            tunnel[pkt_size]["times"],
                            beginning_timestamp,
                            tunnel[pkt_size]["length"],
                            pkt_size,
                        )
                        / tunnel_packets
                    )

                    # determine the case we're measuring against,
                    # if we're the bottom, top or middle of prior measurements
                    (
                        dirty_min,
                        gold_value,
                        dirty_max,
                        dirty_range_max,
                        dirty_range_min,
                    ) = training_percents[pkt_size]

                    # debug(f"calculating '{mine}' distance: ")
                    # debug(f"  {measured=}")
                    # debug(f"  {gold_value=}")
                    # debug(f"  {dirty_min=}")
                    # debug(f"  {dirty_max=}")
                    difference = self.calculate_distance(
                        measured,
                        dirty_min,
                        gold_value,
                        dirty_max,
                        dirty_range_max,
                        dirty_range_min,
                    )

                    difference_total += difference
                    debug(
                        f"{pkt_size}\tm={measured}, percents={training_percents[pkt_size]}\t result: {difference}"
                    )
                    # debug(f"# {spi}\t{seen}\t{signal_importance}\t{deviation}")

                    # XXX: scale result by delta from 0 or from top other
                    # expected signal?

                    # msg = "  {} = seen={}, gold={}, dirt={}, rslt={}"
                    # debug(msg.format(seen,
                    #                  seen_percentage,
                    #                  gold_percent,
                    #                  dirty_percent,
                    #                  result))
                    # msg = "    signal={}, deviation={}"
                    # debug(msg.format(signal_importance, deviation))

                # record the result achieved for this tunnel
                if number_of_sizes > 0:
                    if self.missing_in_tunnel_penalty is not None and len(tunnel) < len(
                        training_percents
                    ):
                        delta = self.missing_in_tunnel_penalty * (
                            len(training_percents) - len(tunnel)
                        )
                        difference_total += delta
                        number_of_sizes += delta
                    average_difference = 1 - (difference_total / number_of_sizes)
                    debug(
                        f"diff total: {protocol_label=} {difference_total=} / {number_of_sizes=} = {average_difference=}"
                    )

                    label = protocol_label
                    relabelled = label
                    if self.label_map and protocol_label in self.label_map:
                        relabelled = self.label_map[protocol_label]

                    # only add it if the new confidence is better than the old
                    if (
                        relabelled in output_results
                        and output_results[relabelled][4] < average_difference
                    ):
                        output_results[relabelled] = (
                            timestamp,
                            identifier,
                            relabelled,
                            average_difference,
                            tunnel_packets,
                        )
                    else:
                        output_results[relabelled] = (
                            timestamp,
                            identifier,
                            relabelled,
                            average_difference,
                            tunnel_packets,
                        )

            # transform the dict results into an array
            output_results = [output_results[x] for x in output_results]

            # output this set of results for this packet for all tunnels
            self.output.output(
                output_results, self.packet_count, len(self.tunnels), self.subscriptions
            )
            output_results = []

        # mark that we're done with a single multi-mine-output
        self.output.end_section()

    def get_packet_identifier(self, pkt):
        identifier = None
        try:
            identifier = (pkt.proto, pkt["IP"].src, pkt["IP"].dst, pkt.spi)
        except Exception:
            pass

        if not identifier:
            if not self.three_tuple_only:
                try:
                    identifier = (
                        pkt.proto,
                        pkt["IP"].src,
                        pkt["IP"].dst,
                        pkt.sport,
                        pkt.dport,
                    )
                except Exception:
                    pass

        if not identifier:
            try:
                identifier = (pkt.proto, pkt["IP"].src, pkt["IP"].dst)
            except Exception:
                identifier = "none"

        return identifier

    def get_dpkt_packet_identifier(self, ip):
        identifier = None

        # ipsec
        if ip.p == 50:
            return (ip.p, "{}.{}.{}.{}".format(*ip.src), "{}.{}.{}.{}".format(*ip.dst))

        if self.three_tuple_only:
            return (ip.p, "{}.{}.{}.{}".format(*ip.src), "{}.{}.{}.{}".format(*ip.dst))

        # udp
        if ip.p == 6 or ip.p == 17:
            udp_tcp = ip.data
            return (
                ip.p,
                "{}.{}.{}.{}".format(*ip.src),
                "{}.{}.{}.{}".format(*ip.dst),
                udp_tcp.sport,
                udp_tcp.dport,
            )

        return (ip.p, "{}.{}.{}.{}".format(*ip.src), "{}.{}.{}.{}".format(*ip.dst))

    def analyze_packet(
        self,
        timestamp,
        pkt,
        window_length: int = 0,
        dig_for_gold: bool = True,
    ) -> None:
        """Analyzes a packet to update the internal model about
        whether its potentially a golden egg and should be prioritized."""

        if self.skip_packets > 0:
            self.skip_packets -= 1
            return

        # get the  headers
        eth = dpkt.ethernet.Ethernet(pkt)
        ip = eth.data
        if not isinstance(ip, dpkt.ip.IP):
            return

        # count the number of packets per SPI
        identifier = self.get_dpkt_packet_identifier(ip)

        # get a unique integer of this index
        if identifier not in self.identifier_indexes:
            self.identifier_indexes[identifier] = len(self.identifier_indexes)
        identifier_index = self.identifier_indexes[identifier]

        pkt_len = len(ip)

        # TODO: this sparse storage is inefficient
        if self.support_windowing:
            self.size_array[
                identifier_index, pkt_len, self.tunnel_packets[identifier]
            ] = True

        # count the total number of packets seen for this identifier
        self.tunnel_packets[identifier] += 1

        # if isinstance(spi, tuple):
        #     spi = "/".join(spi)

        if pkt_len not in self.tunnels[identifier]:
            self.tunnels[identifier][pkt_len] = {
                "length": 0,
                "max_length": self.starting_length,
                "times": np.empty(self.starting_length),
            }

        if self.support_windowing:
            tunnel_size_data = self.tunnels[identifier][pkt_len]
            # spi, length, 'times' is an array of times up to its length 'length' index
            if tunnel_size_data["length"] >= tunnel_size_data["max_length"]:
                warning(
                    f"extending {identifier=} array size to {tunnel_size_data['max_length']*2}"
                )
                tunnel_size_data["times"] = np.hstack(
                    [
                        tunnel_size_data["times"],
                        np.empty(tunnel_size_data["max_length"]),
                    ]
                )
                tunnel_size_data["max_length"] *= 2

            # record the required information for this packet
            tunnel_size_data["times"][tunnel_size_data["length"]] = timestamp
            tunnel_size_data["length"] += 1

        # memorize the last time we see a packet for a given identifier
        last_packet_time[identifier] = timestamp

        # and count the total number of packets
        self.packet_count += 1

        # pass to the main search engine
        if dig_for_gold:
            if (
                self.report_every_n
                and self.tunnel_packets[identifier] % self.report_every_n != 0
            ):
                return
            self.dig_4it(timestamp, window_length, identifier)

    def window_analysis(self):
        # array is:  identifier_number, pkt_size, pkt_sequence_number
        import sys

        oh = pyfsdb.Fsdb(out_file_handle=sys.stdout)
        oh.out_column_names = [
            "label",
            "identifier_num",
            "identifier",
            "window_size",
            "confidence",
        ]

        for mine in self.gold_mines:
            training_percents = self.gold_mines[mine]
            for identifier in self.identifier_indexes:
                identifier_num = self.identifier_indexes[identifier]
                tunnel_pkts = self.tunnel_packets[identifier]

                window_sizes = range(10, max(10000, tunnel_pkts), 10)
                for size in window_sizes:
                    difference = 0.0
                    windows_considered = 0

                    # start rolling the window every 5 packets until the end
                    for offset in range(0, tunnel_pkts - size, 5):

                        # calculate the confidence per packet size
                        for pkt_size in range(1, 1540):
                            if pkt_size not in training_percents:
                                continue
                            windows_considered += (
                                1  # deal with pkt sizes with no training data
                            )

                            (
                                dirty_min,
                                gold_value,
                                dirty_max,
                                dirty_range_max,
                                dirty_range_min,
                            ) = training_percents[pkt_size]

                            # XXX: slide the window
                            window_count = self.size_array[
                                identifier_num, pkt_size, offset : size + offset
                            ].sum()
                            pkt_size_fraction = window_count / size

                            difference += self.calculate_distance(
                                pkt_size_fraction,
                                dirty_min,
                                gold_value,
                                dirty_max,
                                dirty_range_max,
                                dirty_range_min,
                            )

                    debug(f"{windows_considered=} for {identifier=}")
                    if windows_considered == 0:
                        continue
                    difference = difference / windows_considered

                    oh.append([mine, identifier_num, identifier, size, difference])

                # for size in window_sizes:

    def close(self):
        self.output.close()

    def swallow_output(self):
        pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gold-miner",
    version="1.1",
    author="Wes Hardaker",
    author_email="opensource@hardakers.net",
    description="A encrypted tunnel identification research package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/isi-apropos/gold-miner",
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": [
            # migrating to pdb prefixes
            "gold-miner = apropos.goldminer.tools.goldminer:main",
            "gold-miner-trainer = apropos.goldminer.tools.trainer:main",
            "gold-miner-trainer-aggregator = apropos.goldminer.tools.aggregator:main",
            "gold-miner-smelter = apropos.goldminer.tools.smelter:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "pyfsdb",
        "numpy",
        "dpkt",
        "rich",
        "pyaml",
    ],
    python_requires=">=3.6",
    test_suite="nose.collector",
    tests_require=["nose"],
)



#!/usr/bin/python3

"""Scans an interface or pcap file for the likelihood of traffic
within an ipsec/encrypted tunnel for a particular class that you may
want to prioritize."""

import sys
import os

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import collections
import pyfsdb
import numpy as np
import pprint
import logging
from logging import debug, info, warning, error, critical
import time

try:
    import curses
except Exception:
    pass

from apropos.goldminer.pickaxe.pickaxe import PickAxe


def get_parser():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description=__doc__,
        epilog="Exmaple Usage: gold-miner.py -i " + "INTERFACE -g profilename",
    )

    parser.add_argument(
        "-i", "--interface", type=str, help="The interface to monitor for ESP traffic"
    )

    parser.add_argument(
        "-r", "--pcap-file", type=str, help="Read in a PCAP file to analyze"
    )

    parser.add_argument(
        "-p",
        "--training_profile",
        type=str,
        help="The training profile to read for calculating percentages",
    )

    parser.add_argument(
        "-t",
        "--thresholds",
        type=str,
        help="Threshold file to use for determining success",
    )

    parser.add_argument(
        "-j", "--output-json", action="store_true", help="Output data in json format"
    )

    parser.add_argument(
        "-J",
        "--output-flattened-json",
        action="store_true",
        help="Output data in json format, but flattened",
    )

    parser.add_argument(
        "-u", "--output_ui", action="store_true", help="Output data in a window"
    )

    parser.add_argument(
        "-g",
        "--gold-profiles",
        type=str,
        nargs="+",
        help="profiles to identify as 'gold' ; put multiple separated by ,s in an argument",
    )

    parser.add_argument(
        "-a",
        "--all-profiles",
        default=[],
        type=str,
        nargs="*",
        help="Keys to use for all the columns (gold and non-gold)",
    )

    parser.add_argument(
        "--algorithm",
        default="comparison",
        type=str,
        help="Algorithm value to use (lms, linear, comparison)",
    )

    parser.add_argument(
        "-n",
        "--max-packets",
        default=-1,
        type=int,
        help="Maximum number of packets to read",
    )

    parser.add_argument(
        "-N",
        "--report-every",
        default=None,
        type=int,
        help="only report results every N packets",
    )

    parser.add_argument(
        "-L", "--live-results", action="store_true", help="Print live results"
    )

    parser.add_argument(
        "-C",
        "--curses",
        action="store_true",
        help="Turn on a curses view of the output results",
    )

    parser.add_argument(
        "-P",
        "--percentage",
        action="store_true",
        help="Display results as a percentage",
    )

    parser.add_argument(
        "-R",
        "--raw-values",
        action="store_true",
        help="Display raw-value results instead of confidence",
    )

    parser.add_argument(
        "-k",
        "--size-key",
        default="e_pkt_len",
        type=str,
        help="The key to use for pkt size data",
    )

    parser.add_argument(
        "-F",
        "--packet-filter",
        default=None,
        type=str,
        help="Only process these sniffed packets",
    )

    parser.add_argument(
        "-w",
        "--high-low-watermark",
        default=None,
        type=float,
        nargs=2,
        help="Use high/low watermarks to restrict output.  The first argument should be the high value, and the second the low value.",
    )

    parser.add_argument(
        "-3",
        "--three-tuple-only",
        action="store_true",
        help="Only use 3-tuples for analyzing packets instead of 5",
    )

    parser.add_argument(
        "--timing",
        action="store_true",
        help="Add the analysis time length information to the output",
    )

    parser.add_argument(
        "-T",
        "--search-window-length",
        default=None,
        type=float,
        help="Fixed time stamp length to check data over",
    )

    parser.add_argument(
        "-U",
        "--search-window-time-file",
        default=None,
        type=str,
        help="A FSDB file of times to search per packet size",
    )

    parser.add_argument(
        "--log-level",
        "--ll",
        default="info",
        help="Define the logging verbosity level (debug, info, warning, error, fotal, critical).",
    )

    parser.add_argument(
        "--log-file",
        "--lf",
        help="Define a logfile to save logging output to instead of stderr",
    )

    parser.add_argument(
        "--window-analysis",
        action="store_true",
        help="Do window analysis (developer mode)",
    )

    parser.add_argument(
        "output_file",
        type=FileType("w"),
        nargs="?",
        default=sys.stdout,
        help="Where to send the output data to",
    )

    return parser


def parse_args():
    parser = get_parser()
    args = parser.parse_args()

    log_level = args.log_level.upper()
    log_file = args.log_file
    logging.basicConfig(
        level=log_level, filename=log_file, format="%(levelname)-10s:\t%(message)s"
    )

    if not args.gold_profiles or len(args.gold_profiles) == 0:
        raise ValueError("-g is a required parameter")

    if not args.interface and not args.pcap_file:
        raise ValueError("either -i or -r is required")

    return args


def process_arguments(args, axe):
    if args.live_results:
        from apropos.goldminer.output.goldTextDump import GoldTextDump

        axe.output = GoldTextDump()
    elif args.curses:
        from apropos.goldminer.output.goldCurses import GoldCurses

        axe.output = GoldCurses(
            as_percentage=args.percentage, parameter_file=args.thresholds
        )
    elif args.output_json or args.output_flattened_json:
        from apropos.goldminer.output.goldJson import GoldJson

        axe.output = GoldJson(
            as_percentage=args.percentage,
            raw_values=args.raw_values,
            as_list=True,
            parameter_file=args.thresholds,
            flatten=args.output_flattened_json,
        )
    elif args.output_ui:
        try:
            from apropos.goldminer.output.goldQt import GoldQt
        except Exception:
            error("The goldQt module was not found -- please install gold-miner-ui")
            exit(1)

        axe.output = GoldQt()
    else:
        # default to FSDB output formatted
        from apropos.goldminer.output.goldFsdb import GoldFsdb

        axe.output = GoldFsdb(
            out_file_handle=args.output_file,
            as_percentage=args.percentage,
            raw_values=args.raw_values,
            parameter_file=args.thresholds,
        )

    if args.high_low_watermark:
        from apropos.goldminer.output.goldHighLowWatermark import GoldHighLowWatermark

        highlow = GoldHighLowWatermark(
            high=args.high_low_watermark[0],
            low=args.high_low_watermark[1],
            output=axe.output,
        )
        debug("inserted a high-low process")
        axe.output = highlow


def main() -> None:
    """main routine for arg wrapping and calling analysis"""
    args = parse_args()

    # create our searcher
    axe = PickAxe(
        gold_profiles=args.gold_profiles,
        training_data_file=args.training_profile,
        training_key=args.size_key,
        support_windowing=args.window_analysis,
        algorithm=args.algorithm,
        report_every_n=args.report_every,
        three_tuple_only=args.three_tuple_only,
    )

    # read in our training data files, and the keys/etc we expect
    # for both gold and dirty

    process_arguments(args, axe)

    debug("ready to read packets")

    search_window_length = args.search_window_length
    if args.search_window_time_file:
        search_window_length = pyfsdb.Fsdb(
            args.search_window_time_file, return_type=pyfsdb.RETURN_AS_DICTIONARY
        ).get_all()

    run_time_start = time.time()
    if args.interface:
        debug(f"  sniffing({args.interface})")
        axe.process_interface(args.interface, args.packet_filter)
    else:
        (identifiers, counts, processed_data) = axe.process_pcap(
            args.pcap_file,
            None,
            args.max_packets,
            search_window_length,
            dig_for_gold=(not args.window_analysis),
        )
        if args.window_analysis:
            axe.window_analysis()
    run_time_end = time.time()

    if args.timing:
        print(f"# run-time: {run_time_end - run_time_start} seconds")

    info("End of data stream")
    axe.close()


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import sys
import pyfsdb
import collections
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, FileType
import logging
from logging import debug, info, warning, error, critical
from apropos.goldminer.trainer import GoldMineTrainer


def parse_args():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description=__doc__,
        epilog="Exmaple Usage: "
        + "trainer-aggregator.py -o output.fsdb web web.fsdb email email.fsdb",
    )

    parser.add_argument(
        "-o",
        "--output",
        default=sys.stdout,
        type=FileType("w"),
        help="Where to write the output file (else stdout)",
    )

    parser.add_argument(
        "--log-level",
        "--ll",
        default="info",
        help="Define the logging verbosity level (debug, info, warning, error, fotal, critical).",
    )

    parser.add_argument(
        "input_data", nargs="+", help="pairs (label SPACE file ...) to aggregate"
    )

    args = parser.parse_args()

    log_level = args.log_level.upper()
    logging.basicConfig(level=log_level, format="%(levelname)-10s:\t%(message)s")

    if len(args.input_data) % 2 != 0 or len(args.input_data) < 1:
        error("input_file_pairs must be an even number of arguments")

    return args


def aggregate_results(pairs, output):
    results = collections.defaultdict(collections.Counter)
    total_counts = collections.Counter()
    keys = set()
    labels = set()
    converters = {}

    # merge all the data
    for n in range(0, len(pairs), 2):
        (label, filename) = pairs[n : n + 2]
        labels.add(label)
        converters[label] = float

        fh = pyfsdb.Fsdb(filename, return_type=pyfsdb.RETURN_AS_DICTIONARY)
        for row in fh:
            results[label][row["key"]] += row["size_count"]
            total_counts[label] += row["size_count"]
            keys.add(row["key"])

    # make fractions again out of aggregated sizes
    for label in results:
        for key in results[label]:
            results[label][key] /= total_counts[label]

    labels = sorted(labels)

    # output the conglomerates
    oh = pyfsdb.Fsdb(out_file_handle=output, converters=converters)
    oh.out_column_names = ["e_pkt_len"] + labels

    for key in sorted(keys, key=int):
        row = [key] + [results[label][key] for label in labels]
        oh.append(row)

    oh.close()


def main():
    """main routine for arg wrapping and calling analysis"""
    args = parse_args()
    aggregate_results(args.input_data, args.output)


if __name__ == "__main__":
    main()
